from sphinx.util.docfields import Field, TypedField
from sphinx.util.nodes import make_refnode
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain, ObjType, Index
from sphinx.locale import l_, _
from sphinx.roles import XRefRole
from sphinx import addnodes

from docutils.parsers.rst import Directive, directives
from docutils import nodes

class desc_wasp_parameterlist(nodes.Part, nodes.Inline, nodes.TextElement):
  child_text_separator = ' '

class desc_wasp_parameter(nodes.Part, nodes.Inline, nodes.TextElement): pass

def visit_desc_wasp_parameterlist (self, node):
  self.first_param = 1
  self.optional_param_level = 0
  self.required_params_left = sum([isinstance(c, desc_wasp_parameter)
                                  for c in node.children])
  self.param_separator = node.child_text_separator

def depart_desc_wasp_parameterlist (self, node): pass

def visit_desc_wasp_parameter (self, node): self.visit_desc_parameter(node)
def depart_desc_wasp_parameter (self, node): self.depart_desc_parameter(node)

class WaspCommand (ObjectDescription):
  '''
  Description of a command
  '''

  option_spec = {
    'noindex': directives.flag,
    'module': directives.unchanged,
  }

  doc_field_types = [
    TypedField('parameter', label=l_('Parameters'),
               names=('param', 'parameter', 'arg', 'argument'),
               typenames=('type'),
               can_collapse=True),
    Field('return', label=l_('Returns'), names=('returns', 'return')),
  ]

  def get_index_text (self, module, name):
    args = [name]
    if module: args.append(module)
    string = _('{} (in module {})') if module else _('{} (built-in command)')
    return string.format(*args)

  def add_target_and_index(self, name, sig, node):
    module = self.options.get('module', self.env.temp_data.get('wasp:module'))
    fullname = '{}{}'.format((module and '{}|'.format(module) or ''), name)
    if fullname not in self.state.document.ids:
      node['names'].append(fullname)
      node['ids'].append(fullname)
      node['first'] = not self.names
      self.state.document.note_explicit_target(node)
      commands = self.env.domaindata['wasp']['commands']
      if fullname in commands:
        self.state_machine.reporter.warning(
          'duplicate command description of {},\
           other instance in {}, \
           use :noindex for one of them'.format(
             fullname,
             self.env.doc2path(commands[fullname])))
      commands[fullname] = self.env.docname
      index_text = self.get_index_text(module, name)
      if index_text:
        self.indexnode['entries'].append(('single', index_text, fullname, ''))

  def handle_signature (self, sig, node):
    '''
    Transform's a Wasp command signture into RST nodes.
    Returns the command's name

    '''
    # determine module name as well as full name
    signature = sig.split(' ')
    name, arguments = signature[0], signature[1:]
    module = self.options.get('module', self.env.temp_data.get('wasp:module'))
    fullname = name

    node['module'] = module
    node['fullname'] = fullname

    node += addnodes.desc_name(name, name)
    if not arguments: return fullname
    parameter_list = desc_wasp_parameterlist()
    for arg in arguments: parameter_list += desc_wasp_parameter(arg, arg)
    node += parameter_list
    return fullname

class WaspModule(Directive):
  '''
  Directive to mark description of a new module.
  '''
  has_content = False
  required_arguments = 1
  optional_arguments = 0
  final_argument_whitespace = False
  option_spec = {
    'platform': lambda x: x,
    'synopsis': lambda x: x,
    'noindex': directives.flag,
    'deprecated': directives.flag
  }

  def run (self):
    env = self.state.document.settings.env
    name = self.arguments[0].strip()
    noindex = 'noindex' in self.options
    env.temp_data['wasp:module'] = name
    result = []
    if noindex: return result
    value = (env.docname,
             self.options.get('synopsis', ''),
             self.options.get('platform', ''),
             'deprecated' in self.options)
    env.domaindata['wasp']['modules'][name] = value
    long_name = 'module-{}'.format(name)
    target = nodes.target('',
                          '',
                          ids=[long_name],
                          ismod=True)
    self.state.document.note_explicit_target(target)
    result.append(target)
    index_text = _('{} (module)').format(name)
    index_node = addnodes.index(entries=[('single', index_text, long_name, '')])
    result.append(index_node)
    return result


class WaspXRefRole(XRefRole):
  def process_link(self, env, refnode, has_explicit_title, title, target):
    refnode['wasp:module'] = env.temp_data.get('wasp:module')
    exploded = title.split('|')
    title = exploded[1:] and exploded[1:][0] or exploded[0]
    print(title, target)
    return title, target

class WaspModuleIndex(Index): pass

class WaspDomain(Domain):
  '''Wasp language domain'''
  name = 'wasp'
  label = 'Wasp'
  object_types = {
    'command': ObjType(l_('command'), 'command', 'command'),
    'module': ObjType(l_('module'), 'module', 'module')
  }

  directives = {
    'command': WaspCommand,
    'module': WaspModule
  }

  roles = {
    'command': WaspXRefRole(),
    'module': WaspXRefRole()
  }

  initial_data = {
    'commands': { }, # full-name -> docname
    'modules': { } # module-name -> docname, synopsis, platform, deprecated
  }
 # indices = { WaspModuleIndex }

  def clear_doc (self, docname):
    for full, doc in list(self.data['commands'].items()):
      if doc != docname: continue
      del self.data['commands'][full]
    for module, (doc, _, _, _) in list(self.data['modules'].items()):
      if doc != docname: continue
      del self.data['modules'][module]

  def resolve_xref (self, env, from_name, builder, type, target, node, cont):
    module = node.get('wasp:module')
    fullname = '{}{}'.format(module and '{}|'.format(module) or '', target)
    print(fullname)
    path = self.data['commands'].get(fullname)
    if not path: return None
    return make_refnode(builder, from_name, path, fullname, cont, fullname)

  def get_objects (self):
    for module, info in self.data['modules'].items():
      yield (module, module, 'module', info[0], 'module-{}'.format(module), 0)
    for ref, doc in self.data['commands'].items():
      yield (ref, ref, 'command', doc, ref, 1)

def setup (sphinx):
  sphinx.add_node(
    node=desc_wasp_parameterlist,
    html=(visit_desc_wasp_parameterlist, depart_desc_wasp_parameterlist))
  sphinx.add_node(
    node=desc_wasp_parameter,
    html=(visit_desc_wasp_parameter, depart_desc_wasp_parameter))
  sphinx.add_domain(WaspDomain)
